@@ -1,80 +1,60 @@
 from __future__ import annotations
 
-import sys
+from collections.abc import Mapping
 from functools import partial
 from operator import methodcaller
-from typing import TYPE_CHECKING
-from typing import Any
-from typing import Callable
-from typing import Generic
-from typing import Literal
-from typing import Mapping
-from typing import Protocol
-from typing import Sequence
+from typing import TYPE_CHECKING, Any, Callable, Generic, Literal, Protocol
 
-from narwhals._compliant.any_namespace import CatNamespace
-from narwhals._compliant.any_namespace import DateTimeNamespace
-from narwhals._compliant.any_namespace import ListNamespace
-from narwhals._compliant.any_namespace import NameNamespace
-from narwhals._compliant.any_namespace import StringNamespace
-from narwhals._compliant.any_namespace import StructNamespace
+from narwhals._compliant.any_namespace import (
+    CatNamespace,
+    DateTimeNamespace,
+    ListNamespace,
+    NameNamespace,
+    StringNamespace,
+    StructNamespace,
+)
+from narwhals._compliant.column import CompliantColumn
 from narwhals._compliant.namespace import CompliantNamespace
-from narwhals._compliant.typing import AliasName
-from narwhals._compliant.typing import AliasNames
-from narwhals._compliant.typing import CompliantExprT_co
-from narwhals._compliant.typing import CompliantFrameT
-from narwhals._compliant.typing import CompliantLazyFrameT
-from narwhals._compliant.typing import CompliantSeriesOrNativeExprT_co
-from narwhals._compliant.typing import EagerDataFrameT
-from narwhals._compliant.typing import EagerExprT
-from narwhals._compliant.typing import EagerSeriesT
-from narwhals._compliant.typing import LazyExprT
-from narwhals._compliant.typing import NativeExprT
-from narwhals.dependencies import get_numpy
-from narwhals.dependencies import is_numpy_array
-from narwhals.dtypes import DType
-from narwhals.utils import _StoresCompliant
-from narwhals.utils import deprecated
-from narwhals.utils import not_implemented
-
-if not TYPE_CHECKING:  # pragma: no cover
-    if sys.version_info >= (3, 9):
-        from typing import Protocol as Protocol38
-    else:
-        from typing import Generic as Protocol38
-else:  # pragma: no cover
-    # TODO @dangotbanned: Remove after dropping `3.8` (#2084)
-    # - https://github.com/narwhals-dev/narwhals/pull/2064#discussion_r1965921386
-    from typing import Protocol as Protocol38
+from narwhals._compliant.typing import (
+    AliasName,
+    AliasNames,
+    CompliantExprT_co,
+    CompliantFrameT,
+    CompliantLazyFrameT,
+    CompliantSeriesOrNativeExprT_co,
+    EagerDataFrameT,
+    EagerExprT,
+    EagerSeriesT,
+    LazyExprT,
+    NativeExprT,
+)
+from narwhals._utils import _StoresCompliant, qualified_type_name, zip_strict
+from narwhals.dependencies import is_numpy_array, is_numpy_scalar
 
 if TYPE_CHECKING:
-    from typing import Mapping
+    from collections.abc import Mapping, Sequence
 
-    from typing_extensions import Self
-    from typing_extensions import TypeIs
+    from typing_extensions import Self, TypeIs
 
-    from narwhals._compliant.namespace import CompliantNamespace
-    from narwhals._compliant.namespace import EagerNamespace
+    from narwhals._compliant.namespace import CompliantNamespace, EagerNamespace
     from narwhals._compliant.series import CompliantSeries
-    from narwhals._compliant.typing import AliasNames
-    from narwhals._compliant.typing import EvalNames
-    from narwhals._compliant.typing import EvalSeries
-    from narwhals._compliant.typing import ScalarKwargs
-    from narwhals._expression_parsing import ExprKind
-    from narwhals._expression_parsing import ExprMetadata
-    from narwhals.dtypes import DType
-    from narwhals.typing import FillNullStrategy
-    from narwhals.typing import NonNestedLiteral
-    from narwhals.typing import NumericLiteral
-    from narwhals.typing import RankMethod
-    from narwhals.typing import RollingInterpolationMethod
-    from narwhals.typing import TemporalLiteral
-    from narwhals.typing import TimeUnit
-    from narwhals.utils import Implementation
-    from narwhals.utils import Version
-    from narwhals.utils import _FullContext
+    from narwhals._compliant.typing import AliasNames, EvalNames, EvalSeries, ScalarKwargs
+    from narwhals._expression_parsing import ExprKind, ExprMetadata
+    from narwhals._utils import Implementation, Version, _LimitedContext
+    from narwhals.typing import (
+        ClosedInterval,
+        FillNullStrategy,
+        IntoDType,
+        ModeKeepStrategy,
+        NonNestedLiteral,
+        NumericLiteral,
+        RankMethod,
+        RollingInterpolationMethod,
+        TemporalLiteral,
+        TimeUnit,
+    )
 
-__all__ = ["CompliantExpr", "EagerExpr", "LazyExpr", "NativeExpr"]
+__all__ = ["CompliantExpr", "DepthTrackingExpr", "EagerExpr", "LazyExpr", "NativeExpr"]
 
 
 class NativeExpr(Protocol):
@@ -85,13 +65,24 @@ class NativeExpr(Protocol):
     """
 
     def between(self, *args: Any, **kwds: Any) -> Any: ...
-    def isin(self, *args: Any, **kwds: Any) -> Any: ...
+
+    # NOTE: None of these are annotated for `dx.Series`, but are added imperatively
+    # Probably better to define a sub-protocol for `NativeSQLExpr`
+    # - match `dx.Series` to `NativeExpr`
+    # - match the others to `NativeSQLExpr`
+    def __gt__(self, value: Any, /) -> Self: ...
+    def __lt__(self, value: Any, /) -> Self: ...
+    def __ge__(self, value: Any, /) -> Self: ...
+    def __le__(self, value: Any, /) -> Self: ...
+    def __eq__(self, value: Any, /) -> Self: ...  # type: ignore[override]
+    def __ne__(self, value: Any, /) -> Self: ...  # type: ignore[override]
 
 
-class CompliantExpr(Protocol38[CompliantFrameT, CompliantSeriesOrNativeExprT_co]):
+class CompliantExpr(
+    CompliantColumn, Protocol[CompliantFrameT, CompliantSeriesOrNativeExprT_co]
+):
+    # NOTE: `narwhals`
     _implementation: Implementation
-    _backend_version: tuple[int, ...]
-    _version: Version
     _evaluate_output_names: EvalNames[CompliantFrameT]
     _alias_output_names: AliasNames | None
     _metadata: ExprMetadata | None
@@ -99,18 +90,25 @@ class CompliantExpr(Protocol38[CompliantFrameT, CompliantSeriesOrNativeExprT_co]
     def __call__(
         self, df: CompliantFrameT
     ) -> Sequence[CompliantSeriesOrNativeExprT_co]: ...
-    def __narwhals_expr__(self) -> None: ...
+    def __narwhals_expr__(self) -> Self:  # pragma: no cover
+        return self
+
     def __narwhals_namespace__(self) -> CompliantNamespace[CompliantFrameT, Self]: ...
     @classmethod
+    def from_column_indices(
+        cls, *column_indices: int, context: _LimitedContext
+    ) -> Self: ...
+    @classmethod
     def from_column_names(
         cls,
         evaluate_column_names: EvalNames[CompliantFrameT],
         /,
         *,
-        context: _FullContext,
+        context: _LimitedContext,
+    ) -> Self: ...
+    def broadcast(
+        self, kind: Literal[ExprKind.AGGREGATION, ExprKind.LITERAL]
     ) -> Self: ...
-    @classmethod
-    def from_column_indices(cls, *column_indices: int, context: _FullContext) -> Self: ...
     @staticmethod
     def _eval_names_indices(indices: Sequence[int], /) -> EvalNames[CompliantFrameT]:
         def fn(df: CompliantFrameT) -> Sequence[str]:
@@ -119,190 +117,52 @@ class CompliantExpr(Protocol38[CompliantFrameT, CompliantSeriesOrNativeExprT_co]
 
         return fn
 
-    def is_null(self) -> Self: ...
-    def abs(self) -> Self: ...
+    # NOTE: `polars`
     def all(self) -> Self: ...
     def any(self) -> Self: ...
-    def alias(self, name: str) -> Self: ...
-    def cast(self, dtype: DType | type[DType]) -> Self: ...
     def count(self) -> Self: ...
     def min(self) -> Self: ...
     def max(self) -> Self: ...
-    def arg_min(self) -> Self: ...
-    def arg_max(self) -> Self: ...
-    def arg_true(self) -> Self: ...
     def mean(self) -> Self: ...
     def sum(self) -> Self: ...
     def median(self) -> Self: ...
     def skew(self) -> Self: ...
+    def kurtosis(self) -> Self: ...
     def std(self, *, ddof: int) -> Self: ...
     def var(self, *, ddof: int) -> Self: ...
     def n_unique(self) -> Self: ...
     def null_count(self) -> Self: ...
-    def drop_nulls(self) -> Self: ...
-    def fill_null(
-        self,
-        value: Self | NonNestedLiteral,
-        strategy: FillNullStrategy | None,
-        limit: int | None,
-    ) -> Self: ...
-    def diff(self) -> Self: ...
-    def unique(self) -> Self: ...
     def len(self) -> Self: ...
-    def log(self, base: float) -> Self: ...
-    def round(self, decimals: int) -> Self: ...
-    def mode(self) -> Self: ...
-    def head(self, n: int) -> Self: ...
-    def tail(self, n: int) -> Self: ...
-    def shift(self, n: int) -> Self: ...
-    def is_finite(self) -> Self: ...
-    def is_nan(self) -> Self: ...
-    def is_unique(self) -> Self: ...
-    def is_first_distinct(self) -> Self: ...
-    def is_last_distinct(self) -> Self: ...
-    def cum_sum(self, *, reverse: bool) -> Self: ...
-    def cum_count(self, *, reverse: bool) -> Self: ...
-    def cum_min(self, *, reverse: bool) -> Self: ...
-    def cum_max(self, *, reverse: bool) -> Self: ...
-    def cum_prod(self, *, reverse: bool) -> Self: ...
-    def is_in(self, other: Any) -> Self: ...
-    def sort(self, *, descending: bool, nulls_last: bool) -> Self: ...
-    def rank(self, method: RankMethod, *, descending: bool) -> Self: ...
-    def replace_strict(
-        self,
-        old: Sequence[Any] | Mapping[Any, Any],
-        new: Sequence[Any],
-        *,
-        return_dtype: DType | type[DType] | None,
-    ) -> Self: ...
-    def over(
-        self, partition_by: Sequence[str], order_by: Sequence[str] | None
-    ) -> Self: ...
-    def sample(
-        self,
-        n: int | None,
-        *,
-        fraction: float | None,
-        with_replacement: bool,
-        seed: int | None,
-    ) -> Self: ...
+    def over(self, partition_by: Sequence[str], order_by: Sequence[str]) -> Self: ...
     def quantile(
         self, quantile: float, interpolation: RollingInterpolationMethod
     ) -> Self: ...
     def map_batches(
         self,
         function: Callable[[CompliantSeries[Any]], CompliantExpr[Any, Any]],
-        return_dtype: DType | type[DType] | None,
-    ) -> Self: ...
-
-    def clip(
-        self,
-        lower_bound: Self | NumericLiteral | TemporalLiteral | None,
-        upper_bound: Self | NumericLiteral | TemporalLiteral | None,
-    ) -> Self: ...
-
-    def ewm_mean(
-        self,
+        return_dtype: IntoDType | None,
         *,
-        com: float | None,
-        span: float | None,
-        half_life: float | None,
-        alpha: float | None,
-        adjust: bool,
-        min_samples: int,
-        ignore_nulls: bool,
+        returns_scalar: bool,
     ) -> Self: ...
+    @property
+    def name(self) -> NameNamespace[Self]: ...
 
-    def rolling_sum(
-        self,
-        window_size: int,
-        *,
-        min_samples: int,
-        center: bool,
-    ) -> Self: ...
 
-    def rolling_mean(
-        self,
-        window_size: int,
-        *,
-        min_samples: int,
-        center: bool,
-    ) -> Self: ...
-
-    def rolling_var(
-        self,
-        window_size: int,
-        *,
-        min_samples: int,
-        center: bool,
-        ddof: int,
-    ) -> Self: ...
-
-    def rolling_std(
-        self,
-        window_size: int,
-        *,
-        min_samples: int,
-        center: bool,
-        ddof: int,
-    ) -> Self: ...
-
-    @deprecated("Since `1.22.0`")
-    def gather_every(self, n: int, offset: int) -> Self: ...
-    def __and__(self, other: Any) -> Self: ...
-    def __or__(self, other: Any) -> Self: ...
-    def __add__(self, other: Any) -> Self: ...
-    def __sub__(self, other: Any) -> Self: ...
-    def __mul__(self, other: Any) -> Self: ...
-    def __floordiv__(self, other: Any) -> Self: ...
-    def __truediv__(self, other: Any) -> Self: ...
-    def __mod__(self, other: Any) -> Self: ...
-    def __pow__(self, other: Any) -> Self: ...
-    def __gt__(self, other: Any) -> Self: ...
-    def __ge__(self, other: Any) -> Self: ...
-    def __lt__(self, other: Any) -> Self: ...
-    def __le__(self, other: Any) -> Self: ...
-    def __invert__(self) -> Self: ...
-    def broadcast(
-        self, kind: Literal[ExprKind.AGGREGATION, ExprKind.LITERAL]
-    ) -> Self: ...
-    def _is_multi_output_unnamed(self) -> bool:
-        """Return `True` for multi-output aggregations without names.
-
-        For example, column `'a'` only appears in the output as a grouping key:
-
-            df.group_by('a').agg(nw.all().sum())
-
-        It does not get included in:
-
-            nw.all().sum().
-        """
-        assert self._metadata is not None  # noqa: S101
-        return self._metadata.expansion_kind.is_multi_unnamed()
-
-    def _evaluate_aliases(
-        self: CompliantExpr[CompliantFrameT, Any], frame: CompliantFrameT, /
-    ) -> Sequence[str]:
-        names = self._evaluate_output_names(frame)
+class ImplExpr(
+    CompliantExpr[CompliantFrameT, CompliantSeriesOrNativeExprT_co],
+    Protocol[CompliantFrameT, CompliantSeriesOrNativeExprT_co],
+):
+    def _evaluate_aliases(self, frame: CompliantFrameT, /) -> Sequence[str]:
+        # NOTE: Ignore intermittent [False Negative]
+        # Argument of type "CompliantFrameT@ImplExpr" cannot be assigned to parameter of type "CompliantFrameT@ImplExpr"
+        #  Type "CompliantFrameT@ImplExpr" is not assignable to type "CompliantFrameT@ImplExpr"
+        names = self._evaluate_output_names(frame)  # pyright: ignore[reportArgumentType]
         return alias(names) if (alias := self._alias_output_names) else names
 
-    @property
-    def str(self) -> Any: ...
-    @property
-    def name(self) -> Any: ...
-    @property
-    def dt(self) -> Any: ...
-    @property
-    def cat(self) -> Any: ...
-    @property
-    def list(self) -> Any: ...
-    @property
-    def struct(self) -> Any: ...
-
 
 class DepthTrackingExpr(
-    CompliantExpr[CompliantFrameT, CompliantSeriesOrNativeExprT_co],
-    Protocol38[CompliantFrameT, CompliantSeriesOrNativeExprT_co],
+    ImplExpr[CompliantFrameT, CompliantSeriesOrNativeExprT_co],
+    Protocol[CompliantFrameT, CompliantSeriesOrNativeExprT_co],
 ):
     _depth: int
     _function_name: str
@@ -313,7 +173,7 @@ class DepthTrackingExpr(
         evaluate_column_names: EvalNames[CompliantFrameT],
         /,
         *,
-        context: _FullContext,
+        context: _LimitedContext,
         function_name: str = "",
     ) -> Self: ...
 
@@ -340,7 +200,7 @@ class DepthTrackingExpr(
 
 class EagerExpr(
     DepthTrackingExpr[EagerDataFrameT, EagerSeriesT],
-    Protocol38[EagerDataFrameT, EagerSeriesT],
+    Protocol[EagerDataFrameT, EagerSeriesT],
 ):
     _call: EvalSeries[EagerDataFrameT, EagerSeriesT]
     _scalar_kwargs: ScalarKwargs
@@ -354,7 +214,6 @@ class EagerExpr(
         evaluate_output_names: EvalNames[EagerDataFrameT],
         alias_output_names: AliasNames | None,
         implementation: Implementation,
-        backend_version: tuple[int, ...],
         version: Version,
         scalar_kwargs: ScalarKwargs | None = None,
     ) -> None: ...
@@ -365,8 +224,6 @@ class EagerExpr(
     def __narwhals_namespace__(
         self,
     ) -> EagerNamespace[EagerDataFrameT, EagerSeriesT, Self, Any, Any]: ...
-    def __narwhals_expr__(self) -> None: ...
-
     @classmethod
     def _from_callable(
         cls,
@@ -376,7 +233,7 @@ class EagerExpr(
         function_name: str,
         evaluate_output_names: EvalNames[EagerDataFrameT],
         alias_output_names: AliasNames | None,
-        context: _FullContext,
+        context: _LimitedContext,
         scalar_kwargs: ScalarKwargs | None = None,
     ) -> Self:
         return cls(
@@ -386,7 +243,6 @@ class EagerExpr(
             evaluate_output_names=evaluate_output_names,
             alias_output_names=alias_output_names,
             implementation=context._implementation,
-            backend_version=context._backend_version,
             version=context._version,
             scalar_kwargs=scalar_kwargs,
         )
@@ -400,10 +256,47 @@ class EagerExpr(
             evaluate_output_names=lambda _df: [series.name],
             alias_output_names=None,
             implementation=series._implementation,
-            backend_version=series._backend_version,
             version=series._version,
         )
 
+    def _with_alias_output_names(self, alias_name: AliasName | None, /) -> Self:
+        current_alias_output_names = self._alias_output_names
+        alias_output_names: AliasNames | None = (
+            None
+            if alias_name is None
+            else (
+                lambda output_names: [
+                    alias_name(x) for x in current_alias_output_names(output_names)
+                ]
+            )
+            if current_alias_output_names is not None
+            else (lambda output_names: [alias_name(x) for x in output_names])
+        )
+
+        def func(df: EagerDataFrameT) -> list[EagerSeriesT]:
+            if alias_output_names:
+                return [
+                    series.alias(name)
+                    for series, name in zip_strict(
+                        self(df), alias_output_names(self._evaluate_output_names(df))
+                    )
+                ]
+            return [
+                series.alias(name)
+                for series, name in zip_strict(self(df), self._evaluate_output_names(df))
+            ]
+
+        return self.__class__(
+            func,
+            depth=self._depth,
+            function_name=self._function_name,
+            evaluate_output_names=self._evaluate_output_names,
+            alias_output_names=alias_output_names,
+            implementation=self._implementation,
+            version=self._version,
+            scalar_kwargs=self._scalar_kwargs,
+        )
+
     def _reuse_series(
         self,
         method_name: str,
@@ -493,7 +386,7 @@ class EagerExpr(
         self,
         series_namespace: Literal["cat", "dt", "list", "name", "str", "struct"],
         method_name: str,
-        **kwargs: Any,
+        **expressifiable_args: Any,
     ) -> Self:
         """Reuse Series implementation for expression.
 
@@ -503,13 +396,22 @@ class EagerExpr(
         Arguments:
             series_namespace: The Series namespace.
             method_name: name of method, within `series_namespace`.
-            kwargs: keyword arguments to pass to function.
+            expressifiable_args: keyword arguments to pass to function, which may
+                be expressifiable (e.g. `nw.col('a').str.replace('abc', nw.col('b')))`).
         """
-        return self._from_callable(
-            lambda df: [
+
+        def inner(df: EagerDataFrameT) -> list[EagerSeriesT]:
+            kwargs = {
+                name: df._evaluate_expr(value) if self._is_expr(value) else value
+                for name, value in expressifiable_args.items()
+            }
+            return [
                 getattr(getattr(series, series_namespace), method_name)(**kwargs)
                 for series in self(df)
-            ],
+            ]
+
+        return self._from_callable(
+            inner,
             depth=self._depth + 1,
             function_name=f"{self._function_name}->{series_namespace}.{method_name}",
             evaluate_output_names=self._evaluate_output_names,
@@ -535,74 +437,79 @@ class EagerExpr(
             function_name=self._function_name,
             evaluate_output_names=self._evaluate_output_names,
             alias_output_names=self._alias_output_names,
-            backend_version=self._backend_version,
             implementation=self._implementation,
             version=self._version,
             scalar_kwargs=self._scalar_kwargs,
         )
 
-    def cast(self, dtype: DType | type[DType]) -> Self:
+    def cast(self, dtype: IntoDType) -> Self:
         return self._reuse_series("cast", dtype=dtype)
 
+    def _with_binary(self, operator: str, other: Self | Any, /) -> Self:
+        return self._reuse_series(operator, other=other)
+
+    def _with_binary_right(self, operator: str, other: Self | Any, /) -> Self:
+        return self.alias("literal")._reuse_series(operator, other=other)
+
     def __eq__(self, other: Self | Any) -> Self:  # type: ignore[override]
-        return self._reuse_series("__eq__", other=other)
+        return self._with_binary("__eq__", other)
 
     def __ne__(self, other: Self | Any) -> Self:  # type: ignore[override]
-        return self._reuse_series("__ne__", other=other)
+        return self._with_binary("__ne__", other)
 
     def __ge__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__ge__", other=other)
+        return self._with_binary("__ge__", other)
 
     def __gt__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__gt__", other=other)
+        return self._with_binary("__gt__", other)
 
     def __le__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__le__", other=other)
+        return self._with_binary("__le__", other)
 
     def __lt__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__lt__", other=other)
+        return self._with_binary("__lt__", other)
 
     def __and__(self, other: Self | bool | Any) -> Self:
-        return self._reuse_series("__and__", other=other)
+        return self._with_binary("__and__", other)
 
     def __or__(self, other: Self | bool | Any) -> Self:
-        return self._reuse_series("__or__", other=other)
+        return self._with_binary("__or__", other)
 
     def __add__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__add__", other=other)
+        return self._with_binary("__add__", other)
 
     def __sub__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__sub__", other=other)
+        return self._with_binary("__sub__", other)
 
     def __rsub__(self, other: Self | Any) -> Self:
-        return self.alias("literal")._reuse_series("__rsub__", other=other)
+        return self._with_binary_right("__rsub__", other)
 
     def __mul__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__mul__", other=other)
+        return self._with_binary("__mul__", other)
 
     def __truediv__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__truediv__", other=other)
+        return self._with_binary("__truediv__", other)
 
     def __rtruediv__(self, other: Self | Any) -> Self:
-        return self.alias("literal")._reuse_series("__rtruediv__", other=other)
+        return self._with_binary_right("__rtruediv__", other)
 
     def __floordiv__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__floordiv__", other=other)
+        return self._with_binary("__floordiv__", other)
 
     def __rfloordiv__(self, other: Self | Any) -> Self:
-        return self.alias("literal")._reuse_series("__rfloordiv__", other=other)
+        return self._with_binary_right("__rfloordiv__", other)
 
     def __pow__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__pow__", other=other)
+        return self._with_binary("__pow__", other)
 
     def __rpow__(self, other: Self | Any) -> Self:
-        return self.alias("literal")._reuse_series("__rpow__", other=other)
+        return self._with_binary_right("__rpow__", other)
 
     def __mod__(self, other: Self | Any) -> Self:
-        return self._reuse_series("__mod__", other=other)
+        return self._with_binary("__mod__", other)
 
     def __rmod__(self, other: Self | Any) -> Self:
-        return self.alias("literal")._reuse_series("__rmod__", other=other)
+        return self._with_binary_right("__rmod__", other)
 
     # Unary
     def __invert__(self) -> Self:
@@ -640,6 +547,9 @@ class EagerExpr(
     def skew(self) -> Self:
         return self._reuse_series("skew", returns_scalar=True)
 
+    def kurtosis(self) -> Self:
+        return self._reuse_series("kurtosis", returns_scalar=True)
+
     def any(self) -> Self:
         return self._reuse_series("any", returns_scalar=True)
 
@@ -675,6 +585,9 @@ class EagerExpr(
     def is_nan(self) -> Self:
         return self._reuse_series("is_nan")
 
+    def fill_nan(self, value: float | None) -> Self:
+        return self._reuse_series("fill_nan", value=value)
+
     def fill_null(
         self,
         value: Self | NonNestedLiteral,
@@ -682,7 +595,7 @@ class EagerExpr(
         limit: int | None,
     ) -> Self:
         return self._reuse_series(
-            "fill_null", value=value, strategy=strategy, limit=limit
+            "fill_null", value=value, scalar_kwargs={"strategy": strategy, "limit": limit}
         )
 
     def is_in(self, other: Any) -> Self:
@@ -693,7 +606,7 @@ class EagerExpr(
 
     def filter(self, *predicates: Self) -> Self:
         plx = self.__narwhals_namespace__()
-        predicate = plx.all_horizontal(*predicates)
+        predicate = plx.all_horizontal(*predicates, ignore_nulls=False)
         return self._reuse_series("filter", predicate=predicate)
 
     def drop_nulls(self) -> Self:
@@ -704,7 +617,7 @@ class EagerExpr(
         old: Sequence[Any] | Mapping[Any, Any],
         new: Sequence[Any],
         *,
-        return_dtype: DType | type[DType] | None,
+        return_dtype: IntoDType | None,
     ) -> Self:
         return self._reuse_series(
             "replace_strict", old=old, new=new, return_dtype=return_dtype
@@ -749,7 +662,6 @@ class EagerExpr(
             function_name=self._function_name,
             evaluate_output_names=self._evaluate_output_names,
             alias_output_names=alias_output_names,
-            backend_version=self._backend_version,
             implementation=self._implementation,
             version=self._version,
             scalar_kwargs=self._scalar_kwargs,
@@ -769,16 +681,15 @@ class EagerExpr(
     ) -> Self:
         return self._reuse_series(
             "quantile",
-            quantile=quantile,
-            interpolation=interpolation,
             returns_scalar=True,
+            scalar_kwargs={"quantile": quantile, "interpolation": interpolation},
         )
 
     def head(self, n: int) -> Self:
-        return self._reuse_series("head", n=n)
+        return self._reuse_series("head", scalar_kwargs={"n": n})
 
     def tail(self, n: int) -> Self:
-        return self._reuse_series("tail", n=n)
+        return self._reuse_series("tail", scalar_kwargs={"n": n})
 
     def round(self, decimals: int) -> Self:
         return self._reuse_series("round", decimals=decimals)
@@ -789,8 +700,8 @@ class EagerExpr(
     def gather_every(self, n: int, offset: int) -> Self:
         return self._reuse_series("gather_every", n=n, offset=offset)
 
-    def mode(self) -> Self:
-        return self._reuse_series("mode")
+    def mode(self, *, keep: ModeKeepStrategy) -> Self:
+        return self._reuse_series("mode", scalar_kwargs={"keep": keep})
 
     def is_finite(self) -> Self:
         return self._reuse_series("is_finite")
@@ -798,9 +709,11 @@ class EagerExpr(
     def rolling_mean(self, window_size: int, *, min_samples: int, center: bool) -> Self:
         return self._reuse_series(
             "rolling_mean",
-            window_size=window_size,
-            min_samples=min_samples,
-            center=center,
+            scalar_kwargs={
+                "window_size": window_size,
+                "min_samples": min_samples,
+                "center": center,
+            },
         )
 
     def rolling_std(
@@ -808,15 +721,22 @@ class EagerExpr(
     ) -> Self:
         return self._reuse_series(
             "rolling_std",
-            window_size=window_size,
-            min_samples=min_samples,
-            center=center,
-            ddof=ddof,
+            scalar_kwargs={
+                "window_size": window_size,
+                "min_samples": min_samples,
+                "center": center,
+                "ddof": ddof,
+            },
         )
 
     def rolling_sum(self, window_size: int, *, min_samples: int, center: bool) -> Self:
         return self._reuse_series(
-            "rolling_sum", window_size=window_size, min_samples=min_samples, center=center
+            "rolling_sum",
+            scalar_kwargs={
+                "window_size": window_size,
+                "min_samples": min_samples,
+                "center": center,
+            },
         )
 
     def rolling_var(
@@ -824,33 +744,51 @@ class EagerExpr(
     ) -> Self:
         return self._reuse_series(
             "rolling_var",
-            window_size=window_size,
-            min_samples=min_samples,
-            center=center,
-            ddof=ddof,
+            scalar_kwargs={
+                "window_size": window_size,
+                "min_samples": min_samples,
+                "center": center,
+                "ddof": ddof,
+            },
         )
 
     def map_batches(
         self,
         function: Callable[[Any], Any],
-        return_dtype: DType | type[DType] | None,
+        return_dtype: IntoDType | None,
+        *,
+        returns_scalar: bool,
     ) -> Self:
         def func(df: EagerDataFrameT) -> Sequence[EagerSeriesT]:
-            input_series_list = self(df)
-            output_names = [input_series.name for input_series in input_series_list]
-            result = [function(series) for series in input_series_list]
-            if is_numpy_array(result[0]) or (
-                (np := get_numpy()) is not None and np.isscalar(result[0])
-            ):
-                from_numpy = partial(
-                    self.__narwhals_namespace__()._series.from_numpy, context=self
-                )
-                result = [
-                    from_numpy(array).alias(output_name)
-                    for array, output_name in zip(result, output_names)
-                ]
+            udf_series_in = self(df)
+            output_names = (input_series.name for input_series in udf_series_in)
+            udf_series_out = tuple(function(series) for series in udf_series_in)
+            _first_in, _first_out = udf_series_in[0], udf_series_out[0]
+
+            result: Sequence[EagerSeriesT]
+            it = zip_strict(udf_series_out, output_names)
+            if is_numpy_array(_first_out) or is_numpy_scalar(_first_out):
+                from_numpy = partial(_first_in.from_numpy, context=self)
+                result = tuple(from_numpy(arr).alias(out_name) for arr, out_name in it)
+            elif isinstance(_first_out, _first_in.__class__):  # compliant series
+                result = tuple(series.alias(out_name) for series, out_name in it)
+            else:  # If everything else fails, assume scalar case
+                from_scalar = _first_in._from_scalar
+                result = tuple(from_scalar(val).alias(out_name) for val, out_name in it)
+
             if return_dtype is not None:
-                result = [series.cast(return_dtype) for series in result]
+                result = tuple(series.cast(return_dtype) for series in result)
+
+            is_scalar_result = tuple(len(r) == 1 for r in result)
+            if (not returns_scalar) and any(is_scalar_result) and (len(df) > 1):
+                _idx = is_scalar_result.index(True)  # Index of first result with length 1
+                _type = type(udf_series_out[_idx])
+                msg = (
+                    "`map_batches` with `returns_scalar=False` must return a Series; "
+                    f"found '{qualified_type_name(_type)}'.\n\nIf `returns_scalar` "
+                    "is set to `True`, a returned value can be a scalar value."
+                )
+                raise TypeError(msg)
             return result
 
         return self._from_callable(
@@ -862,6 +800,61 @@ class EagerExpr(
             context=self,
         )
 
+    def shift(self, n: int) -> Self:
+        return self._reuse_series("shift", scalar_kwargs={"n": n})
+
+    def cum_sum(self, *, reverse: bool) -> Self:
+        return self._reuse_series("cum_sum", scalar_kwargs={"reverse": reverse})
+
+    def cum_count(self, *, reverse: bool) -> Self:
+        return self._reuse_series("cum_count", scalar_kwargs={"reverse": reverse})
+
+    def cum_min(self, *, reverse: bool) -> Self:
+        return self._reuse_series("cum_min", scalar_kwargs={"reverse": reverse})
+
+    def cum_max(self, *, reverse: bool) -> Self:
+        return self._reuse_series("cum_max", scalar_kwargs={"reverse": reverse})
+
+    def cum_prod(self, *, reverse: bool) -> Self:
+        return self._reuse_series("cum_prod", scalar_kwargs={"reverse": reverse})
+
+    def rank(self, method: RankMethod, *, descending: bool) -> Self:
+        return self._reuse_series(
+            "rank", scalar_kwargs={"method": method, "descending": descending}
+        )
+
+    def log(self, base: float) -> Self:
+        return self._reuse_series("log", base=base)
+
+    def exp(self) -> Self:
+        return self._reuse_series("exp")
+
+    def sqrt(self) -> Self:
+        return self._reuse_series("sqrt")
+
+    def is_between(
+        self, lower_bound: Any, upper_bound: Any, closed: ClosedInterval
+    ) -> Self:
+        return self._reuse_series(
+            "is_between", lower_bound=lower_bound, upper_bound=upper_bound, closed=closed
+        )
+
+    def is_close(
+        self,
+        other: Self | NumericLiteral,
+        *,
+        abs_tol: float,
+        rel_tol: float,
+        nans_equal: bool,
+    ) -> Self:
+        return self._reuse_series(
+            "is_close",
+            other=other,
+            abs_tol=abs_tol,
+            rel_tol=rel_tol,
+            nans_equal=nans_equal,
+        )
+
     @property
     def cat(self) -> EagerExprCatNamespace[Self]:
         return EagerExprCatNamespace(self)
@@ -887,29 +880,10 @@ class EagerExpr(
         return EagerExprStructNamespace(self)
 
 
-class LazyExpr(
-    CompliantExpr[CompliantLazyFrameT, NativeExprT],
-    Protocol38[CompliantLazyFrameT, NativeExprT],
+# mypy thinks `NativeExprT` should be covariant, pyright thinks it should be invariant
+class LazyExpr(  # type: ignore[misc]
+    ImplExpr[CompliantLazyFrameT, NativeExprT], Protocol[CompliantLazyFrameT, NativeExprT]
 ):
-    arg_min: not_implemented = not_implemented()
-    arg_max: not_implemented = not_implemented()
-    arg_true: not_implemented = not_implemented()
-    head: not_implemented = not_implemented()
-    tail: not_implemented = not_implemented()
-    mode: not_implemented = not_implemented()
-    sort: not_implemented = not_implemented()
-    sample: not_implemented = not_implemented()
-    map_batches: not_implemented = not_implemented()
-    ewm_mean: not_implemented = not_implemented()
-    gather_every: not_implemented = not_implemented()
-    replace_strict: not_implemented = not_implemented()
-    cat: not_implemented = not_implemented()  # pyright: ignore[reportAssignmentType]
-
-    @classmethod
-    def _is_expr(cls, obj: Self | Any) -> TypeIs[Self]:
-        return hasattr(obj, "__narwhals_expr__")
-
-    def _with_callable(self, call: Callable[..., Any], /) -> Self: ...
     def _with_alias_output_names(self, func: AliasNames | None, /) -> Self: ...
     def alias(self, name: str) -> Self:
         def fn(names: Sequence[str]) -> Sequence[str]:
@@ -920,9 +894,6 @@ class LazyExpr(
 
         return self._with_alias_output_names(fn)
 
-    @classmethod
-    def _alias_native(cls, expr: NativeExprT, name: str, /) -> NativeExprT: ...
-
     @property
     def name(self) -> LazyExprNameNamespace[Self]:
         return LazyExprNameNamespace(self)
@@ -1030,6 +1001,9 @@ class EagerExprDateTimeNamespace(
     def truncate(self, every: str) -> EagerExprT:
         return self.compliant._reuse_series_namespace("dt", "truncate", every=every)
 
+    def offset_by(self, by: str) -> EagerExprT:
+        return self.compliant._reuse_series_namespace("dt", "offset_by", by=by)
+
 
 class EagerExprListNamespace(
     EagerExprNamespace[EagerExprT], ListNamespace[EagerExprT], Generic[EagerExprT]
@@ -1037,6 +1011,15 @@ class EagerExprListNamespace(
     def len(self) -> EagerExprT:
         return self.compliant._reuse_series_namespace("list", "len")
 
+    def unique(self) -> EagerExprT:
+        return self.compliant._reuse_series_namespace("list", "unique")
+
+    def contains(self, item: NonNestedLiteral) -> EagerExprT:
+        return self.compliant._reuse_series_namespace("list", "contains", item=item)
+
+    def get(self, index: int) -> EagerExprT:
+        return self.compliant._reuse_series_namespace("list", "get", index=index)
+
 
 class CompliantExprNameNamespace(  # type: ignore[misc]
     _ExprNamespace[CompliantExprT_co],
@@ -1044,7 +1027,7 @@ class CompliantExprNameNamespace(  # type: ignore[misc]
     Protocol[CompliantExprT_co],
 ):
     def keep(self) -> CompliantExprT_co:
-        return self._from_callable(lambda name: name, alias=False)
+        return self._from_callable(None)
 
     def map(self, function: AliasName) -> CompliantExprT_co:
         return self._from_callable(function)
@@ -1068,9 +1051,7 @@ class CompliantExprNameNamespace(  # type: ignore[misc]
 
         return fn
 
-    def _from_callable(
-        self, func: AliasName, /, *, alias: bool = True
-    ) -> CompliantExprT_co: ...
+    def _from_callable(self, func: AliasName | None, /) -> CompliantExprT_co: ...
 
 
 class EagerExprNameNamespace(
@@ -1078,22 +1059,9 @@ class EagerExprNameNamespace(
     CompliantExprNameNamespace[EagerExprT],
     Generic[EagerExprT],
 ):
-    def _from_callable(self, func: AliasName, /, *, alias: bool = True) -> EagerExprT:
+    def _from_callable(self, func: AliasName | None) -> EagerExprT:
         expr = self.compliant
-        return type(expr)(
-            lambda df: [
-                series.alias(func(name))
-                for series, name in zip(expr(df), expr._evaluate_output_names(df))
-            ],
-            depth=expr._depth,
-            function_name=expr._function_name,
-            evaluate_output_names=expr._evaluate_output_names,
-            alias_output_names=self._alias_output_names(func) if alias else None,
-            backend_version=expr._backend_version,
-            implementation=expr._implementation,
-            version=expr._version,
-            scalar_kwargs=expr._scalar_kwargs,
-        )
+        return expr._with_alias_output_names(func)
 
 
 class LazyExprNameNamespace(
@@ -1101,9 +1069,9 @@ class LazyExprNameNamespace(
     CompliantExprNameNamespace[LazyExprT],
     Generic[LazyExprT],
 ):
-    def _from_callable(self, func: AliasName, /, *, alias: bool = True) -> LazyExprT:
+    def _from_callable(self, func: AliasName | None) -> LazyExprT:
         expr = self.compliant
-        output_names = self._alias_output_names(func) if alias else None
+        output_names = self._alias_output_names(func) if func else None
         return expr._with_alias_output_names(output_names)
 
 
@@ -1150,12 +1118,18 @@ class EagerExprStringNamespace(
     def to_datetime(self, format: str | None) -> EagerExprT:
         return self.compliant._reuse_series_namespace("str", "to_datetime", format=format)
 
+    def to_date(self, format: str | None) -> EagerExprT:
+        return self.compliant._reuse_series_namespace("str", "to_date", format=format)
+
     def to_lowercase(self) -> EagerExprT:
         return self.compliant._reuse_series_namespace("str", "to_lowercase")
 
     def to_uppercase(self) -> EagerExprT:
         return self.compliant._reuse_series_namespace("str", "to_uppercase")
 
+    def zfill(self, width: int) -> EagerExprT:
+        return self.compliant._reuse_series_namespace("str", "zfill", width=width)
+
 
 class EagerExprStructNamespace(
     EagerExprNamespace[EagerExprT], StructNamespace[EagerExprT], Generic[EagerExprT]
